"""
import_to_cpg.py 
Imports directories of pictures to picture albums in a coppermine 
gallery database.

MYSQL_HOST    : mysql server to connect to
MYSQL_USER    : username used to log into mysql server
MYSQL_PW      : password used to log into mysql server
MYSQL_DB      : database name to add items to / remove items from
SRC_DIR       : coppermine image directory
EXCLUDE_DIRS  : directories to ignore
ADD           : true to add categories, albums and images, false to remove them
TRACE_QUERIES : true displays all queries submitted to mysql database
VERBOSE       : true displays verbose information  
ADD_IMAGES    : true adds images - default true
ADD_VIDEO     : true adds video files - default false
ADD_AUDIO     : true adds audio files - default false

COPPERMINE_PREFIX           : prefix of all coppermine tables
COPPERMINE_ALBUMS_TABLE     : coppermine table of albums
COPPERMINE_PICTURES_TABLE   : coppermine table of pictures
COPPERMINE_CATEGORIES_TABLE : coppermine table of categories

ALBUM_DESCRIPTION_FILENAME : 
text file located in each directory of images which is assigned 
to the description of the album created from the directory.  If no
text file is present then no description is assigned.

IMAGE_NORMAL_WIDTH_HEIGHT : normal dimensions of an image in the gallery 
IMAGE_THUMB_WIDTH_HEIGHT  : thumbnail dimensions of an image in the gallery
IMAGE_JPG_QUALITY         : resize image quality 0...100
"""

import os
import sys
import getopt
import stat
import time
import MySQLdb
import thumbs

MYSQL_HOST    = 'localhost'
MYSQL_USER    = 'root'
MYSQL_PW      = ''
MYSQL_DB      = 'coppermine'

SRC_DIR       = '.'
EXCLUDE_DIRS  = []
ADD           = True

TRACE_QUERIES = False
VERBOSE       = True

ADD_IMAGES    = True
ADD_VIDEO     = False
ADD_AUDIO     = False

COPPERMINE_PREFIX           = 'cpg1411_'
COPPERMINE_ALBUMS_TABLE     = COPPERMINE_PREFIX + 'albums'
COPPERMINE_PICTURES_TABLE   = COPPERMINE_PREFIX + 'pictures'
COPPERMINE_CATEGORIES_TABLE = COPPERMINE_PREFIX + 'categories'
COPPERMINE_VIDEO_EXTENSIONS = [ '.mov',
                                '.avi' ]
COPPERMINE_AUDIO_EXTENSIONS = [ '.wav',
                                '.mp3' ]
COPPERMINE_IMAGE_EXTENSIONS = [ '.jpg',
                                '.gif',
                                '.bmp',
                                '.tga',
                                '.png' ]

ALBUM_DESCRIPTION_FILENAME  = 'description.txt'

IMAGE_NORMAL_WIDTH_HEIGHT   = 400
IMAGE_THUMB_WIDTH_HEIGHT    = 100
IMAGE_JPG_QUALITY           = 85
IMAGE_NORMAL_PREFIX         = 'normal_'
IMAGE_THUMB_PREFIX          = 'thumb_'

# This logic can be customized to build an category name
# from the album path.
# This rule takes the first two digits of a directory
# containing images and builds a year category
def category_from_album_path(albumPath):
    """Build a category from an albumPath"""
    while True:
        path = os.path.split(albumPath)
        if not path[1]:
            break
        try:
            yrstr = path[1][:2]
            year  = int(yrstr)
            if year < 50:
                category = '20' + yrstr
            else:
                category = '19' + yrstr
            return category
        except:
            pass
        albumPath = path[0]

    return 'Miscellaneous'


def character_repeat(char, count):
    """Build a string by repeating the specified character count times"""
    s = ''
    for i in xrange(0, count):
        s = s + char
    return s


def sql_execute(query, cursor):
    """SQL a query on a mysql database using the specified cursor"""
    trace = TRACE_QUERIES
    if trace:
        print query

    cursor.execute(query)
    
    results = []
    while 1:
        row = cursor.fetchone()
        if row == None:
            break

        results.append(row[0])
        if trace:
            print row[0]

    return results


def sql_select(table, columns, whereExpression, cursor):
    """Build a SQL query to select columns from the specified table"""
    query = 'SELECT ('
    for r in xrange(0, len(columns)):
        query = query + str(columns[r])
        if r == len(columns) - 1:
            query = query + ') FROM ' + table + ' WHERE '
        else:
            query = query + ', '

    query = query + whereExpression
    return sql_execute(query, cursor)


def sql_delete(table, whereExpression, cursor):
    """Build a SQL query to delete rows from a table"""
    query = 'DELETE FROM ' + table + ' WHERE ' + whereExpression
    return sql_execute(query, cursor)


def sql_insert(table, keyValues, cursor):
    """Builds a SQL query to onsert keyValues into the table using 
    specified mysql cursor"""
    query = 'INSERT INTO ' + table + ' ('
    for k in xrange(0, len(keyValues)):
        query = query + str(keyValues.keys()[k])
        if k == len(keyValues.keys()) - 1:
            query = query + ') VALUES '
        else:
            query = query + ', '

    numRows = 0
    for k in xrange(0, len(keyValues)):
        numRows = max(numRows, len(keyValues.values()[k]))

    for r in xrange(0, numRows):
        query = query + '('
        for k in xrange(0, len(keyValues)):
            query = query + str(keyValues.values()[k][r])
            if k < len(keyValues) - 1:
                query = query + ', '
        query = query + ')'
        if r < numRows - 1:
            query = query + ', '

    return sql_execute(query, cursor)


def relative_path(path, relativeDir):
    """Returns path relative to relativeDir"""
    # use normpath to ensure path separators are uniform
    path = os.path.normpath(path)
    prelen = len(os.path.commonprefix((
                os.path.normcase(path),
                # add a separator to get correct prefix length
                # (normpath removes trailing separators)
                os.path.normcase(os.path.normpath(relativeDir)) + 
                os.sep
                )))
    return path[prelen:]


def collate_albums(srcdir, album = None, albums = []):
    """Build a list of albums with contained pictures from directories 
    containing images under srcdir"""
    # get file list
    for f in os.listdir(srcdir):
        fullpath = os.path.join(srcdir, f)
        filestat = os.stat(fullpath)
        # if this is a directory add album
        if stat.S_ISDIR(filestat.st_mode):
            # unless the directory has been excluded
            addAlbum = True
            for exclude in EXCLUDE_DIRS:
                if exclude in fullpath:
                    addAlbum = False
                    break

            if not addAlbum:
                continue

            # album name is relative to the base directory
            if album:
                f = album['name'] + '/' + f

            # TODO : make the category customizable from command line
            category = category_from_album_path(fullpath)

            # read description file
            descriptionFilename = os.path.join(fullpath, 
                                               ALBUM_DESCRIPTION_FILENAME)
            albumDescription = ""
            if os.path.exists(descriptionFilename):
                try:
                    descriptionFile  = open(descriptionFilename)
                    albumDescription = descriptionFile.read(-1)
                    descriptionFile.close()
                except IOError:
                    print 'Warning unable to read ' + descriptionFilename

            albums.append({'name'         : f, 
                           'path'         : fullpath,
                           'stat'         : filestat, 
                           'categoryName' : category,
                           'category'     : None,
                           'images'       : [],
                           'description'  : albumDescription})
            collate_albums(fullpath, albums[len(albums) - 1], albums)
        # if we have an album and this is an image add to the album
        elif album and stat.S_ISREG(filestat.st_mode):
            # don't add thumb or 'normal' images
            if not ((IMAGE_NORMAL_PREFIX in f) or
                    (IMAGE_THUMB_PREFIX  in f)):
                album['images'].append({'name' : f,
                                        'path' : fullpath,
                                        'stat' : filestat})
    return albums


def collate_categories(albums):
    """Build dictionary of categories referenced by albums 
    containing more than one image"""
    categories = {}
    for a in albums:
        if len(a['images']):
            try:
                category = categories[a['categoryName']]
            except KeyError:
                categories[a['categoryName']] = {
                    'name' : a['categoryName'],
                    'id'   : None}
                category = categories[a['categoryName']]
            a['category'] = category
    return categories


def main():
    try:
        opts, args = getopt.getopt(sys.argv[1:], 's:e:ru:p:d:vn:t:q:maP?')
    except getopt.GetoptError, v:
        print v
        opts = (('-?', '-?'), )

    # process command line options
    for opt in opts:
        # strip -
        arg = opt[0][1:]
        if arg == 's':
            if os.path.exists(opt[1]):
                globals()['SRC_DIR'] = opt[1]
        elif arg == 'e':
            globals()['EXCLUDE_DIRS'] = opt[1].split()
        elif arg == 'r':
            globals()['ADD'] = False
        elif arg == 'h':
            globals()['MYSQL_HOST'] = opt[1]
        elif arg == 'u':
            globals()['MYSQL_USER'] = opt[1]
        elif arg == 'p':
            globals()['MYSQL_PW'] = opt[1]
        elif arg == 'd':
            globals()['MYSQL_DB'] = opt[1]
        elif arg == 'n':
            globals()['IMAGE_NORMAL_WIDTH_HEIGHT'] = int(opt[1])
        elif arg == 't':
            globals()['IMAGE_THUMB_WIDTH_HEIGHT']  = int(opt[1])
        elif arg == 'q':
            globals()['IMAGE_JPG_QUALITY'] = opt[1]
        elif arg == 'v':
            globals()['TRACE_QUERIES'] = True
            globals()['VERBOSE']       = True
        elif arg == 'm':
            globals()['ADD_VIDEO'] = True
        elif arg == 'a':
            globals()['ADD_AUDIO'] = True
        elif arg == 'P':
            globals()['ADD_IMAGES'] = False
        elif arg == '?':
            # display help text
            print 'python import_to_cpg.py {switches}'
            print 'all of switches below are optional:'
            print '-s source_directory          : path to search for photos'
            print '-e exclude_directory_pattern : pattern which excludes '
            print '                               directories'
            print '-r                           : remove processed pictures'
            print '                               and albums from database'
            print '-u mysql_username            : username to log into mysql'
            print '-p mysql_password            : password to log into mysql'
            print '-d mysql_database_name       : name of mysql database to use'
            print '-v                           : trace database queries'
            print '-n widthxheight              : normal picture dimensions'
            print '-t widthxheight              : thumb picture dimensions'
            print '-q jpg_quality               : jpg image quality'
            print '-m                           : add movie /video files'
            print '-a                           : add audio files'
            print '-P                           : don\'t add images'
            print '-?                           : display this help text'
            print 
            print 'example:'
            print 'python import_to_cpy.py -d=c:/test -u=root -p=password \\'
            print '                        -d=pictures'
            sys.exit(-1)

    # display variables which control the app
    print 'Database connection info'
    print 'MYSQL_HOST   = ' + MYSQL_HOST
    print 'MYSQL_USER   = ' + MYSQL_USER
    print 'MYSQL_PW     = ' + character_repeat('*', len(MYSQL_PW))
    print 'MYSQL_DB     = ' + MYSQL_DB
    print 
    print 'Image acquisition info'
    print 'SRC_DIR      = ' + SRC_DIR
    print 'EXCLUDE_DIRS = ' + str(EXCLUDE_DIRS)
    print 'ADD          = ' + str(ADD)
    print 'ADD_IMAGES   = ' + str(ADD_IMAGES)
    print 'ADD_VIDEO    = ' + str(ADD_VIDEO)
    print 'ADD_AUDIO    = ' + str(ADD_AUDIO)
    print
    print 'IMAGE_NORMAL_WIDTH_HEIGHT = ' + str(IMAGE_NORMAL_WIDTH_HEIGHT)
    print 'IMAGE_THUMB_WIDTH_HEIGHT  = ' + str(IMAGE_THUMB_WIDTH_HEIGHT)
    print 'IMAGE_JPG_QUALITY         = ' + str(IMAGE_JPG_QUALITY)

    # open db
    conn = MySQLdb.connect(
        host = MYSQL_HOST, 
        user = MYSQL_USER, 
        passwd = MYSQL_PW,
        db = MYSQL_DB)
    cursor = conn.cursor()

    # build list of albums and categories
    os.chdir(SRC_DIR)
    globals()['SRC_DIR'] = os.getcwd()
    albums     = collate_albums(SRC_DIR)
    categories = collate_categories(albums)

    for c in categories.values():
        # remove category
        sql_delete(COPPERMINE_CATEGORIES_TABLE, 
                   'name = ' + conn.string_literal(c['name']), cursor)
        if ADD:
            # add category and get ID
            rows = {'name'        : [conn.string_literal(c['name']),],
                    'description' : ['""',]}
            sql_insert(COPPERMINE_CATEGORIES_TABLE, rows, cursor)
            c['id'] = sql_select(
                COPPERMINE_CATEGORIES_TABLE, ['cid'], 
                'name = ' + conn.string_literal(c['name']),
                cursor)[0]


    for a in albums:
        # skip empty albums
        if 0 == len(a['images']):
            continue

        # remove album 
        sql_delete(COPPERMINE_ALBUMS_TABLE, 
                   'title = "' + a['name'] + '"', cursor)
        if ADD:
            # add album
            rows = {'title'       : [conn.string_literal(a['name']), ],
                    'category'    : [conn.string_literal(
                        str(a['category']['id'])), ],
                    'description' : [conn.string_literal(a['description']), ]}
            sql_insert(COPPERMINE_ALBUMS_TABLE, rows, cursor)
            a['id'] = sql_select(
                COPPERMINE_ALBUMS_TABLE, ['aid', ],
                'title = "' + a['name'] + '"', cursor)[0]

        
        # add images                
        for i in a['images']:
            # build normal and thumbnail paths
            imagePath  = os.path.split(i['path'])
            fileExt    = (os.path.splitext(imagePath[1])[1]).lower()
            avFile     = ((ADD_VIDEO and 
                           (fileExt in COPPERMINE_VIDEO_EXTENSIONS)) or
                          (ADD_AUDIO and 
                           (fileExt in COPPERMINE_AUDIO_EXTENSIONS)))
            normalPath = os.path.join(imagePath[0], 
                                      IMAGE_NORMAL_PREFIX + imagePath[1])
            thumbPath  = os.path.join(imagePath[0], 
                                      IMAGE_THUMB_PREFIX + imagePath[1])
            # remove entry from db
            sql_delete(COPPERMINE_PICTURES_TABLE,
                       'filename = "' + i['name'] + '"', cursor)
            if ADD:
                thumbInfo  = None
                normalInfo = None

                # create 'normal' image and thumbnail
                if (ADD_IMAGES and 
                    (fileExt in COPPERMINE_IMAGE_EXTENSIONS)):
                    thumbInfo = thumbs.resize(
                        i['path'], thumbPath,
                        IMAGE_THUMB_WIDTH_HEIGHT,
                        IMAGE_JPG_QUALITY, VERBOSE)
                    normalInfo = thumbs.resize(
                        i['path'], normalPath,
                        IMAGE_NORMAL_WIDTH_HEIGHT,
                        IMAGE_JPG_QUALITY, VERBOSE)

                if avFile:
                  normalInfo = {'srcwh' : [0, 0]}
                # if we failed to create either normal or thumb, 
                # or this isn't a audio/video file, clean up and skip 
                elif not (thumbInfo or normalInfo):
                    if thumbInfo:
                        os.remove(thumbPath)
                    if normalInfo:
                        os.remove(normalPath)
                    continue
                
                # add image to db
                rows = {
                    'aid'            : [str(a['id']), ],
                    'filepath'       : [conn.string_literal(
                            os.path.split(
                                relative_path(
                                    i['path'], SRC_DIR)
                                )[0].replace('\\', '/') + '/'), ],
                    'filename'       : [conn.string_literal(
                            i['name']), ],
                    'filesize'       : [str(i['stat'].st_size), ],
                    'total_filesize' : [str(i['stat'].st_size), ],
                    'pwidth'         : [str(normalInfo['srcwh'][0]), ],
                    'pheight'        : [str(normalInfo['srcwh'][1]), ],
                    'mtime'          : [conn.string_literal(
                            time.strftime(
                                '%Y-%m-%d %H:%M:%S', 
                                time.gmtime(
                                    i['stat'].st_mtime))), ],
                    'ctime'          : [i['stat'].st_ctime, ],
                    'title'          : [conn.string_literal(
                            os.path.splitext(
                                i['name'])[0]), ],
                    'caption'        : ['""', ],
                    'keywords'       : ['""', ],
                    'user1'          : ['""', ],
                    'user2'          : ['""', ],
                    'user3'          : ['""', ],
                    'user4'          : ['""', ],
                    'approved'       : ['1', ], 
                    'owner_id'       : ['1', ],
                    'owner_name'     : ['"admin"', ], }
                sql_insert(COPPERMINE_PICTURES_TABLE, rows, cursor)
            else:
                # remove thumb and normal image
                if os.path.exists(thumbPath):
                    if VERBOSE:
                        print 'deleting ' + thumbPath
                    os.remove(thumbPath)
                if os.path.exists(normalPath):
                    if VERBOSE:
                        print 'deleting ' + normalPath  
                    os.remove(normalPath)

    # clean up
    cursor.close()
    conn.close()

if __name__ == '__main__':
    main();
